#ifndef _CRT_SECURE_NO_WARNINGS
#define _CRT_SECURE_NO_WARNINGS
#endif
#ifndef _WIN32_WINNT
#define _WIN32_WINNT 0x0601
#endif
#ifndef NTDDI_VERSION
#define NTDDI_VERSION NTDDI_WIN7
#endif
#ifndef WIN32_LEAN_AND_MEAN
#define WIN32_LEAN_AND_MEAN
#endif
#ifndef _DLL
#define _DLL
#endif
#ifndef _UNICODE
#define _UNICODE
#endif
#ifndef UNICODE
#define UNICODE
#endif

#include <winstrct.h>
#include <objbase.h>
#include <shellapi.h>
#include <devioctl.h>
#include <ntddscsi.h>
#include <ntdddisk.h>
#include <..\ddk\mountmgr.h>
#include <winioctl.h>
#include <ntdll.h>
#include <stdlib.h>

#pragma comment(lib, "shell32.lib")
#pragma comment(lib, "ole32.lib")
#pragma comment(lib, "msvcrt.lib")
#pragma comment(lib, "ntdllp.lib")

#define MAX_NUMBER_OF_DISK_EXTENTS 20
#define VOLUME_DISK_EXTENTS_BYTE_SIZE (sizeof(VOLUME_DISK_EXTENTS) + MAX_NUMBER_OF_DISK_EXTENTS * sizeof(DISK_EXTENT))

#pragma pack(push)
#pragma pack(1)
typedef struct _FAT_BPB
{
    WORD BytesPerSector;
    BYTE SectorsPerCluster;
    WORD ReservedSectors;
    BYTE NumberOfFileAllocationTables;
    WORD NumberOfRootEntries;
    WORD NumberOfSectors;
    BYTE MediaDescriptor;
    WORD SectorsPerFileAllocationTable;
    WORD SectorsPerTrack;
    WORD NumberOfHeads;
    union
    {
        struct
        {
            WORD NumberOfHiddenSectors;
            WORD TotalNumberOfSectors;
        } DOS_320;
        struct
        {
            DWORD NumberOfHiddenSectors;
            DWORD TotalNumberOfSectors;
        } DOS_331;
    };
} FAT_BPB, *PFAT_BPB;

typedef struct _EXT_BPB
{
    BYTE PhysicalDriveNumber;
    BYTE CurrentHead;
    BYTE Signature;
    DWORD Id;
    CHAR VolumeLabel[11];
    CHAR SystemId[8];
} EXT_BPB, *PEXT_BPB;

typedef struct _FAT_VBR
{
    BYTE JumpInstruction[3];
    CHAR OEMName[8];
    FAT_BPB BPB;
    EXT_BPB ExtendedBPB;
    BYTE FillData1[512 -
        3 -
        8 -
        sizeof(FAT_BPB) -
        sizeof(EXT_BPB) -
        2];
    BYTE Signature[2];
} FAT_VBR, *PFAT_VBR;
#pragma pack(pop)

LPSTR
CleanNonPrintables(LPSTR str,
int size)
{
    int i;
    for (i = 0; i < size; i++)
        if ((str[i] & 0xE0) == 0)
            str[i] = 0x20;

    return str;
}

POBJECT_NAME_INFORMATION
GetNativeTargetName(LPCWSTR name);

int
wmain(int argc, LPWSTR *argv)
{
    DWORD dwIoControlCode = 0;
    BOOL displayhelp = FALSE;
    HANDLE hDev;
    WCHAR cBuf[MAX_PATH] = L"\\\\?\\";
    DWORD dw;
    
    if (argc < 3)
        displayhelp = TRUE;
    else if (!_wcsicmp(argv[1], L"DISMOUNT"))
        dwIoControlCode = FSCTL_LOCK_VOLUME;
    else if (!_wcsicmp(argv[1], L"FORCEDISMOUNT"))
        dwIoControlCode = FSCTL_DISMOUNT_VOLUME;
    else if (!_wcsicmp(argv[1], L"FLUSH"))
        dwIoControlCode = 0;
    else if (!_wcsicmp(argv[1], L"LOAD"))
        dwIoControlCode = IOCTL_STORAGE_LOAD_MEDIA;
    else if (!_wcsicmp(argv[1], L"EJECT"))
        dwIoControlCode = IOCTL_STORAGE_EJECT_MEDIA;
    else if (!_wcsicmp(argv[1], L"GEOMETRY"))
        dwIoControlCode = IOCTL_DISK_GET_DRIVE_GEOMETRY;
    else if (!_wcsicmp(argv[1], L"SCSIADDRESS"))
        dwIoControlCode = IOCTL_SCSI_GET_ADDRESS;
    else if (!_wcsicmp(argv[1], L"EXTEND"))
        dwIoControlCode = FSCTL_EXTEND_VOLUME;
    else if (!_wcsicmp(argv[1], L"GROW"))
        dwIoControlCode = IOCTL_DISK_GROW_PARTITION;
    else if (!_wcsicmp(argv[1], L"UPDATE"))
        dwIoControlCode = IOCTL_DISK_UPDATE_PROPERTIES;
    else if (!_wcsicmp(argv[1], L"OFFLINE"))
        dwIoControlCode = IOCTL_VOLUME_OFFLINE;
    else if (!_wcsicmp(argv[1], L"ONLINE"))
        dwIoControlCode = IOCTL_VOLUME_ONLINE;
    else if (!_wcsicmp(argv[1], L"ATTRIB"))
        dwIoControlCode = IOCTL_DISK_SET_DISK_ATTRIBUTES;
    else if (!_wcsicmp(argv[1], L"EXTENTS"))
        dwIoControlCode = IOCTL_VOLUME_GET_VOLUME_DISK_EXTENTS;
    else if (!_wcsicmp(argv[1], L"SETPARTTYPE"))
        dwIoControlCode = IOCTL_DISK_SET_PARTITION_INFO;
    else if (!_wcsicmp(argv[1], L"ASSIGN"))
        dwIoControlCode = IOCTL_MOUNTMGR_NEXT_DRIVE_LETTER;
    else if (!_wcsicmp(argv[1], L"RESCAN"))
        dwIoControlCode = IOCTL_SCSI_RESCAN_BUS;
    else
        displayhelp = TRUE;

    if (displayhelp)
    {
        puts("Command line interface to Win32 API function\r\n"
            "DeviceIoControl.\r\n"
            "\n"
            "Usage:\r\n"
            "DEVIOCTL command device\r\n"
            "\n"
            "Command can be either of:\n"
            "DISMOUNT      Dismount filesystem unless in use.\r\n"
            "FORCEDISMOUNT Forcefully dismount filesystem.\r\n"
            "FLUSH         Flush filesystem buffers.\r\n"
            "LOAD          Load CD/DVD media.\r\n"
            "EJECT         Eject CD/DVD media or removable devices.\r\n"
            "GEOMETRY      Display geometry and partition type information.\r\n"
            "SCSIADDRESS   Display SCSI address for device.\r\n"
            "RESCAN        Rescan bus on specified storage port.\r\n"
            "UPDATE        Update disk properties.\r\n"
            "OFFLINE       Bring disk volume offline.\r\n"
            "ONLINE        Bring disk volume online.\r\n"
            "EXTENTS       Display volume disk extents.\r\n"
            "ASSIGN        Assign a drive letter to a disk volume.\r\n"
            "\n"
            "Device can be a drive letter followed by a colon or Volume{GUID} syntax to\r\n"
            "specify a disk volume, PhysicalDriveN syntax to specify a physical disk drive,\r\n"
            "COMn syntax to specify a serial port, ScsiN: syntax to specify storage ports\r\n"
            "and so on.\r\n"
            "\n"
            "Other operations:\n"
            "\n"
            "DEVIOCTL EXTEND d: totalnew\r\n"
            "\"totalnew\" is total new size of filesystem in sectors.\r\n"
            "d: is a disk volume, e.g. C:.\r\n"
            "\n"
            "DEVIOCTL GROW PhysicalDriveN partno growbytes\r\n"
            "\"partno\" is the partition number.\r\n"
            "\"growbytes\" is the number of bytes to grow.\r\n"
            "\n"
            "DEVIOCTL SETPARTTYPE d: parttype\r\n"
            "d: is a disk volume, e.g. C:.\r\n"
            "parttype is new partition type id, e.g. 7 for Windows IFS partition type.\r\n"
            "\n"
            "DEVIOCTL ATTRIB PhysicalDriveN [PERSIST] [READONLY|WRITEABLE] [ONLINE|OFFLINE]\r\n"
            "PhysicalDriveN is physical hard drive, e.g. PhysicalDrive2.\r\n"
            "PERSIST saves changes until next time disk is attached or next reboot.");

        return -1;
    }

    SetErrorMode(SEM_FAILCRITICALERRORS);
    SetProcessShutdownParameters(0x3FF, SHUTDOWN_NORETRY);

    ++argv;
    --argc;

    wcsncpy(cBuf + 4, argv[1], _countof(cBuf) - 4);
    cBuf[_countof(cBuf) - 1] = 0;

    hDev = INVALID_HANDLE_VALUE;
    switch (dwIoControlCode)
    {
    case IOCTL_SCSI_RESCAN_BUS:
        hDev = CreateFile(cBuf, GENERIC_READ | GENERIC_WRITE,
            FILE_SHARE_READ | FILE_SHARE_WRITE, NULL,
            OPEN_EXISTING, 0, NULL);

        if (hDev == INVALID_HANDLE_VALUE)
            break;

        DeviceIoControl(hDev,
            IOCTL_SCSI_RESCAN_BUS,
            NULL,
            0,
            NULL,
            0,
            &dw,
            NULL);

        break;

    case IOCTL_DISK_GET_DRIVE_GEOMETRY:
    {
        static union
        {
            FAT_VBR fat_vbr;
            char buffer[65536];
        } vbr = { 0 };

        static DISK_GEOMETRY disk_geometry = { 0 };
        static PARTITION_INFORMATION partition_information = { 0 };
        static PARTITION_INFORMATION_EX partition_information_ex = { 0 };
        static GET_LENGTH_INFORMATION length_information = { 0 };
        static DWORDLONG total_size = 0;
        static STORAGE_DEVICE_NUMBER storage_device_number = { 0 };

        hDev = CreateFile(cBuf, GENERIC_READ,
            FILE_SHARE_READ | FILE_SHARE_WRITE, NULL,
            OPEN_EXISTING, 0, NULL);
        if (hDev == INVALID_HANDLE_VALUE)
            break;

        DeviceIoControl(hDev,
            IOCTL_DISK_UPDATE_PROPERTIES,
            NULL,
            0,
            NULL,
            0,
            &dw,
            NULL);

        if (DeviceIoControl(hDev,
            IOCTL_DISK_GET_DRIVE_GEOMETRY,
            NULL,
            0,
            &disk_geometry,
            sizeof disk_geometry,
            &dw,
            NULL))
        {
            total_size = disk_geometry.Cylinders.QuadPart *
                disk_geometry.TracksPerCylinder *
                disk_geometry.SectorsPerTrack *
                disk_geometry.BytesPerSector;

            printf("Disk CHS geometry:\n"
                "Media type    : 0x%.2X\n"
                "Cylinders     : %I64u\n"
                "Tracks per cyl: %u\n"
                "Sectors per tr: %u\n"
                "Bytes per sect: %u\n"
                "Total CHS size: %I64u (%.4g %s)\n",
                disk_geometry.MediaType,
                disk_geometry.Cylinders.QuadPart,
                disk_geometry.TracksPerCylinder,
                disk_geometry.SectorsPerTrack,
                disk_geometry.BytesPerSector,
                total_size, TO_h(total_size), TO_p(total_size));
        }
        else
            win_perror(L"IOCTL_DISK_GET_DRIVE_GEOMETRY");

        if (DeviceIoControl(hDev,
            IOCTL_DISK_GET_LENGTH_INFO,
            NULL,
            0,
            &length_information,
            sizeof length_information,
            &dw,
            NULL))
            printf("\n"
            "LBA length    : %I64u (%.4g %s)\n",
            length_information.Length.QuadPart,
            TO_h(length_information.Length.QuadPart),
            TO_p(length_information.Length.QuadPart));

        if (DeviceIoControl(hDev,
            IOCTL_DISK_GET_PARTITION_INFO_EX,
            NULL,
            0,
            &partition_information_ex,
            sizeof partition_information_ex,
            &dw,
            NULL))
        {
            printf("\n"
                "Partition information:\n"
                "Start offset  : %I64u\n"
                "Partition size: %I64u (%.4g %s)\n"
                "Partition no  : %u\n",
                partition_information_ex.StartingOffset.QuadPart,
                partition_information_ex.PartitionLength.QuadPart,
                TO_h(partition_information_ex.PartitionLength.QuadPart),
                TO_p(partition_information_ex.PartitionLength.QuadPart),
                partition_information_ex.PartitionNumber);

            switch (partition_information_ex.PartitionStyle)
            {
            case PARTITION_STYLE_MBR:
                printf(
                    "Partition tbl : MBR\n"
                    "Hidden sectors: %u\n"
                    "Partition type: 0x%.2X\n"
                    "Boot indicator: %u\n"
                    "Recognized par: %u\n",
                    partition_information_ex.Mbr.HiddenSectors,
                    (DWORD)partition_information_ex.Mbr.PartitionType,
                    partition_information_ex.Mbr.BootIndicator,
                    partition_information_ex.Mbr.RecognizedPartition);

                break;

            case PARTITION_STYLE_GPT:
            {
                static OLECHAR type_guid_chars[40] = { 0 };
                static OLECHAR id_guid_chars[40] = { 0 };

                (void)StringFromGUID2(&partition_information_ex.Gpt.PartitionType,
                    type_guid_chars, _countof(type_guid_chars));

                (void)StringFromGUID2(&partition_information_ex.Gpt.PartitionId,
                    id_guid_chars, _countof(id_guid_chars));

                printf(
                    "Partition tbl : GPT\n"
                    "Attributes    : 0x%.16I64X\n"
                    "Partition type: %ws\n"
                    "Partition id  : %ws\n"
                    "Partition name: '%.*ws'\n",
                    partition_information_ex.Gpt.Attributes,
                    type_guid_chars,
                    id_guid_chars,
                    _countof(partition_information_ex.Gpt.Name),
                    partition_information_ex.Gpt.Name);

                break;
            }

            default:
                break;
            }
        }
        else if (DeviceIoControl(hDev,
            IOCTL_DISK_GET_PARTITION_INFO_EX,
            NULL,
            0,
            &partition_information,
            sizeof partition_information,
            &dw,
            NULL))
        {
            printf("\n"
                "Partition information:\n"
                "Start offset  : %I64u\n"
                "Partition size: %I64u (%.4g %s)\n"
                "Hidden sectors: %u\n"
                "Partition no  : %u\n"
                "Partition type: 0x%.2X\n"
                "Boot indicator: %u\n"
                "Recognized par: %u\n",
                partition_information.StartingOffset.QuadPart,
                partition_information.PartitionLength.QuadPart,
                TO_h(partition_information.PartitionLength.QuadPart),
                TO_p(partition_information.PartitionLength.QuadPart),
                partition_information.HiddenSectors,
                partition_information.PartitionNumber,
                (DWORD)partition_information.PartitionType,
                partition_information.BootIndicator,
                partition_information.RecognizedPartition);
        }

        if (DeviceIoControl(hDev,
            IOCTL_STORAGE_GET_DEVICE_NUMBER,
            NULL,
            0,
            &storage_device_number,
            sizeof storage_device_number,
            &dw,
            NULL))
            printf("\n"
            "Device number information:\n"
            "Device type   : 0x%.2X\n"
            "Device number : %i\n"
            "Partition no  : %i\n",
            storage_device_number.DeviceType,
            storage_device_number.DeviceNumber,
            storage_device_number.PartitionNumber);

        if (ReadFile(hDev, &vbr, sizeof(vbr), &dw, NULL))
            printf("\n"
            "Volume boot record values:\n"
            "OEM name      : %.*s\n"
            "Bytes per sect: %u\n"
            "Sectors per cl: %u\n"
            "Reserved sect : %u\n"
            "FAT count     : %u\n"
            "FAT root entr : %u\n"
            "Total sectors : %u\n"
            "Media descript: 0x%.2X\n"
            "Sectors pr FAT: %u\n"
            "Sect per track: %u\n"
            "Number of head: %u\n"
            "Physicl drv no: 0x%.2X\n"
            "Signature     : 0x%.2X\n"
            "Id            : 0x%.8X\n"
            "Volume label  : %.*s\n"
            "System Id     : %.*s\n",
            sizeof(vbr.fat_vbr.OEMName),
            CleanNonPrintables(vbr.fat_vbr.OEMName,
            sizeof(vbr.fat_vbr.OEMName)),
            (UINT)vbr.fat_vbr.BPB.BytesPerSector,
            (UINT)vbr.fat_vbr.BPB.SectorsPerCluster,
            (UINT)vbr.fat_vbr.BPB.ReservedSectors,
            (UINT)vbr.fat_vbr.BPB.NumberOfFileAllocationTables,
            (UINT)vbr.fat_vbr.BPB.NumberOfRootEntries,
            (UINT)vbr.fat_vbr.BPB.NumberOfSectors,
            (INT)vbr.fat_vbr.BPB.MediaDescriptor,
            (UINT)vbr.fat_vbr.BPB.SectorsPerFileAllocationTable,
            (UINT)vbr.fat_vbr.BPB.SectorsPerTrack,
            (UINT)vbr.fat_vbr.BPB.NumberOfHeads,
            (UINT)vbr.fat_vbr.ExtendedBPB.PhysicalDriveNumber,
            (UINT)vbr.fat_vbr.ExtendedBPB.Signature,
            (UINT)vbr.fat_vbr.ExtendedBPB.Id,
            sizeof(vbr.fat_vbr.ExtendedBPB.VolumeLabel),
            CleanNonPrintables
            (vbr.fat_vbr.ExtendedBPB.VolumeLabel,
            sizeof(vbr.fat_vbr.ExtendedBPB.VolumeLabel)),
            sizeof(vbr.fat_vbr.ExtendedBPB.SystemId),
            CleanNonPrintables(vbr.fat_vbr.ExtendedBPB.SystemId,
            sizeof(vbr.fat_vbr.ExtendedBPB.SystemId)));
        else
            win_perror(L"Error reading first sector");

        SetLastError(NO_ERROR);
        break;
    }

    case IOCTL_SCSI_GET_ADDRESS:
    {
        static SCSI_ADDRESS scsi_address = { 0 };
        scsi_address.Length = sizeof(scsi_address);

        hDev = CreateFile(cBuf, GENERIC_READ,
            FILE_SHARE_READ | FILE_SHARE_WRITE, NULL,
            OPEN_EXISTING, 0, NULL);
        if (hDev == INVALID_HANDLE_VALUE)
            break;

        DeviceIoControl(hDev,
            IOCTL_DISK_UPDATE_PROPERTIES,
            NULL,
            0,
            NULL,
            0,
            &dw,
            NULL);

        if (!DeviceIoControl(hDev,
            IOCTL_SCSI_GET_ADDRESS,
            NULL,
            0,
            &scsi_address,
            sizeof scsi_address,
            &dw,
            NULL))
            break;

        printf("Disk SCSI address:\n"
            "Port number   : %i\n"
            "Path id       : %i\n"
            "Target id     : %i\n"
            "Logical unit n: %i\n",
            (int)scsi_address.PortNumber,
            (int)scsi_address.PathId,
            (int)scsi_address.TargetId,
            (int)scsi_address.Lun);


        SetLastError(NO_ERROR);
        break;
    }

    case IOCTL_VOLUME_GET_VOLUME_DISK_EXTENTS:
    {
        PDISK_EXTENT extent;
        static union
        {
            VOLUME_DISK_EXTENTS disk_extents;
            char buffer[VOLUME_DISK_EXTENTS_BYTE_SIZE];
        } extents;

        hDev = CreateFile(cBuf, GENERIC_READ,
            FILE_SHARE_READ | FILE_SHARE_WRITE, NULL,
            OPEN_EXISTING, 0, NULL);
        if (hDev == INVALID_HANDLE_VALUE)
            break;

        if (!DeviceIoControl(hDev,
            IOCTL_VOLUME_GET_VOLUME_DISK_EXTENTS,
            NULL,
            0,
            &extents,
            VOLUME_DISK_EXTENTS_BYTE_SIZE,
            &dw,
            NULL))
            break;

        for (extent = extents.disk_extents.Extents;
            extents.disk_extents.NumberOfDiskExtents--;
            extent++)
        {
            printf("Disk number    : %u\n"
                "Starting offset: %I64i\n"
                "Extent length  : %I64i (%.4g %s)\n"
                "\n",
                extent->DiskNumber,
                extent->StartingOffset.QuadPart,
                extent->ExtentLength.QuadPart,
                TO_h(extent->ExtentLength.QuadPart),
                TO_p(extent->ExtentLength.QuadPart));
        }

        SetLastError(NO_ERROR);
        break;
    }

    case IOCTL_DISK_SET_DISK_ATTRIBUTES:
    {
        DWORD required_access = GENERIC_READ;
        GET_DISK_ATTRIBUTES get_disk_attributes = { sizeof get_disk_attributes };
        SET_DISK_ATTRIBUTES set_disk_attributes = { sizeof set_disk_attributes };

        while (argc >= 3)
        {
            if (_wcsicmp(argv[2], L"PERSIST") == 0)
            {
                set_disk_attributes.Persist = TRUE;
            }
            else if (_wcsicmp(argv[2], L"ONLINE") == 0)
            {
                set_disk_attributes.AttributesMask |= DISK_ATTRIBUTE_OFFLINE;
                set_disk_attributes.Attributes &= ~DISK_ATTRIBUTE_OFFLINE;
            }
            else if (_wcsicmp(argv[2], L"OFFLINE") == 0)
            {
                set_disk_attributes.AttributesMask |= DISK_ATTRIBUTE_OFFLINE;
                set_disk_attributes.Attributes |= DISK_ATTRIBUTE_OFFLINE;
            }
            else if (_wcsicmp(argv[2], L"READONLY") == 0)
            {
                set_disk_attributes.AttributesMask |= DISK_ATTRIBUTE_READ_ONLY;
                set_disk_attributes.Attributes |= DISK_ATTRIBUTE_READ_ONLY;
            }
            else if (_wcsicmp(argv[2], L"WRITEABLE") == 0)
            {
                set_disk_attributes.AttributesMask |= DISK_ATTRIBUTE_READ_ONLY;
                set_disk_attributes.Attributes &= ~DISK_ATTRIBUTE_READ_ONLY;
            }

            argv++;
            argc--;
        }

        if (set_disk_attributes.AttributesMask != 0)
        {
            required_access |= GENERIC_WRITE;
        }

        hDev = CreateFile(cBuf, required_access,
            FILE_SHARE_READ | FILE_SHARE_WRITE, NULL,
            OPEN_EXISTING, 0, NULL);
        if (hDev == INVALID_HANDLE_VALUE)
            break;

        if (set_disk_attributes.AttributesMask != 0)
            if (!DeviceIoControl(hDev,
                IOCTL_DISK_SET_DISK_ATTRIBUTES,
                &set_disk_attributes,
                sizeof set_disk_attributes,
                NULL,
                0,
                &dw,
                NULL))
                break;

        if (!DeviceIoControl(hDev,
            IOCTL_DISK_GET_DISK_ATTRIBUTES,
            NULL,
            0,
            &get_disk_attributes,
            sizeof get_disk_attributes,
            &dw,
            NULL))
            break;

        printf("Disk attributes:\n"
            "%s\n"
            "%s\n",
            get_disk_attributes.Attributes & DISK_ATTRIBUTE_OFFLINE ? "OFFLINE" : "ONLINE",
            get_disk_attributes.Attributes & DISK_ATTRIBUTE_READ_ONLY ? "READONLY" : "WRITEABLE");

        SetLastError(NO_ERROR);
        break;
    }

    case IOCTL_VOLUME_OFFLINE:
    case IOCTL_VOLUME_ONLINE:
    case FSCTL_LOCK_VOLUME:
    case IOCTL_STORAGE_EJECT_MEDIA:
        hDev = CreateFile(cBuf, GENERIC_READ | GENERIC_WRITE,
            FILE_SHARE_READ | FILE_SHARE_WRITE, NULL,
            OPEN_EXISTING, 0, NULL);

        if (hDev == INVALID_HANDLE_VALUE)
            hDev = CreateFile(cBuf, FILE_READ_ATTRIBUTES | SYNCHRONIZE,
            FILE_SHARE_READ | FILE_SHARE_WRITE, NULL,
            OPEN_EXISTING, 0, NULL);

        if (hDev == INVALID_HANDLE_VALUE)
            break;

        FlushFileBuffers(hDev);

        if (!DeviceIoControl(hDev, FSCTL_LOCK_VOLUME, NULL, 0, NULL, 0, &dw,
            NULL))
            break;

    case FSCTL_DISMOUNT_VOLUME:
        if (hDev == INVALID_HANDLE_VALUE)
            hDev = CreateFile(cBuf, GENERIC_READ | GENERIC_WRITE,
            FILE_SHARE_READ | FILE_SHARE_WRITE, NULL,
            OPEN_EXISTING, 0, NULL);
        if (hDev == INVALID_HANDLE_VALUE)
            hDev = CreateFile(cBuf, FILE_READ_ATTRIBUTES | SYNCHRONIZE,
            FILE_SHARE_READ | FILE_SHARE_WRITE, NULL,
            OPEN_EXISTING, 0, NULL);
        if (hDev == INVALID_HANDLE_VALUE)
            break;

        if (!DeviceIoControl(hDev, FSCTL_DISMOUNT_VOLUME, NULL, 0, NULL, 0,
            &dw, NULL))
            break;

        switch (dwIoControlCode)
        {
        case IOCTL_STORAGE_EJECT_MEDIA:
        case IOCTL_VOLUME_OFFLINE:
        case IOCTL_VOLUME_ONLINE:
        {
            if (DeviceIoControl(hDev, dwIoControlCode, NULL, 0,
                NULL, 0, &dw, NULL))
                SetLastError(NO_ERROR);

            break;
        }

        default:
            SetLastError(NO_ERROR);
        }

        break;

    case IOCTL_DISK_UPDATE_PROPERTIES:
    case IOCTL_STORAGE_LOAD_MEDIA:
        hDev = CreateFile(cBuf, FILE_READ_ATTRIBUTES | SYNCHRONIZE,
            FILE_SHARE_READ | FILE_SHARE_WRITE, NULL,
            OPEN_EXISTING, 0, NULL);
        if (hDev == INVALID_HANDLE_VALUE)
            break;

        if (DeviceIoControl(hDev, dwIoControlCode, NULL, 0, NULL, 0,
            &dw, NULL))
            SetLastError(NO_ERROR);

        break;

    case IOCTL_DISK_SET_PARTITION_INFO:
    {
        SET_PARTITION_INFORMATION part_info;
        WCHAR *endptr = NULL;
        long value = 0;

        if (argc < 3)
        {
            fputs("Function requires an argument.\n", stderr);
            return 1;
        }
            
        value = wcstol(argv[2], &endptr, 0);

        if (value < 0 || value > MAXBYTE ||
            (endptr != NULL && *endptr != 0))
        {
            fprintf(stderr, "Invalid partition type: %ws\n", argv[2]);
            return 1;
        }

        part_info.PartitionType = (BYTE)value;

        hDev = CreateFile(cBuf, GENERIC_READ | GENERIC_WRITE,
            FILE_SHARE_READ | FILE_SHARE_WRITE, NULL,
            OPEN_EXISTING, 0, NULL);
        if (hDev == INVALID_HANDLE_VALUE)
            break;

        if (DeviceIoControl(hDev, IOCTL_DISK_SET_PARTITION_INFO, &part_info,
            sizeof part_info, NULL, 0, &dw, NULL))
            SetLastError(NO_ERROR);

        break;
    }

    case IOCTL_MOUNTMGR_NEXT_DRIVE_LETTER:
    {
        PMOUNTMGR_DRIVE_LETTER_TARGET target;
        DWORD target_size;
        MOUNTMGR_DRIVE_LETTER_INFORMATION info;
        POBJECT_NAME_INFORMATION object_name = GetNativeTargetName(cBuf);
        if (object_name == NULL)
            break;

        hDev = CreateFileW(MOUNTMGR_DOS_DEVICE_NAME,
            GENERIC_READ | GENERIC_WRITE,
            FILE_SHARE_READ | FILE_SHARE_WRITE,
            NULL, OPEN_EXISTING, 0, NULL);

        if (hDev == INVALID_HANDLE_VALUE)
            break;

        target_size = FIELD_OFFSET(MOUNTMGR_DRIVE_LETTER_TARGET, DeviceName) + object_name->Name.Length;
        target = HeapAlloc(GetProcessHeap(), HEAP_GENERATE_EXCEPTIONS, target_size);

        __analysis_assume(target != NULL);

        target->DeviceNameLength = object_name->Name.Length;
        memcpy(target->DeviceName, object_name->Name.Buffer, object_name->Name.Length);

        HeapFree(GetProcessHeap(), 0, object_name);

        if (DeviceIoControl(hDev, IOCTL_MOUNTMGR_NEXT_DRIVE_LETTER,
            target, target_size,
            &info, sizeof info,
            &dw, NULL))
        {
            printf(
                "Assigned    : %s\n"
                "Drive letter: %c\n",
                info.DriveLetterWasAssigned ? "Yes" : "No",
                info.CurrentDriveLetter);

            SetLastError(NO_ERROR);
        }

        HeapFree(GetProcessHeap(), 0, target);

        break;
    }

    case FSCTL_EXTEND_VOLUME:
    {
        LONGLONG newsize;
        WCHAR suffix;

        if (argc < 3)
        {
            fputs("Function requires an argument.\n", stderr);
            return 1;
        }

        switch (swscanf(argv[2], L"%I64i%c",
            &newsize, &suffix))
        {
        case 2:
            switch (suffix)
            {
            case L'G':
                newsize <<= 10;
            case L'M':
                newsize <<= 10;
            case L'K':
                newsize <<= 10;
                break;

            default:
                fprintf(stderr, "Invalid size suffix: %wc\n", suffix);
                return 1;
            }

        case 1:
            break;

        default:
            fprintf(stderr, "Invalid grow size: %ws\n", argv[3]);
            return 1;
        }

        hDev = CreateFile(cBuf, GENERIC_READ | GENERIC_WRITE,
            FILE_SHARE_READ | FILE_SHARE_WRITE, NULL,
            OPEN_EXISTING, 0, NULL);
        if (hDev == INVALID_HANDLE_VALUE)
            break;

        if (DeviceIoControl(hDev, FSCTL_EXTEND_VOLUME, &newsize,
            sizeof newsize, NULL, 0, &dw, NULL))
            SetLastError(NO_ERROR);

        break;
    }

    case IOCTL_DISK_GROW_PARTITION:
    {
        DISK_GROW_PARTITION disk_grow_partition = { 0 };
        WCHAR suffix;

        if (argc < 4)
        {
            fputs("Function requires two arguments.\n", stderr);
            return 1;
        }

        if (swscanf(argv[2], L"%i", &disk_grow_partition.PartitionNumber) <
            1)
        {
            fprintf(stderr, "Invalid partition number: %ws\n", argv[2]);
            return 1;
        }

        switch (swscanf(argv[3], L"%I64i%c",
            &disk_grow_partition.BytesToGrow, &suffix))
        {
        case 2:
            switch (suffix)
            {
            case L'G':
                disk_grow_partition.BytesToGrow.QuadPart <<= 10;
            case L'M':
                disk_grow_partition.BytesToGrow.QuadPart <<= 10;
            case L'K':
                disk_grow_partition.BytesToGrow.QuadPart <<= 10;
                break;

            default:
                fprintf(stderr, "Invalid size suffix: %wc\n", suffix);
                return 1;
            }

        case 1:
            break;

        default:
            fprintf(stderr, "Invalid grow size: %ws\n", argv[3]);
            return 1;
        }

        hDev = CreateFile(cBuf, GENERIC_READ | GENERIC_WRITE,
            FILE_SHARE_READ | FILE_SHARE_WRITE, NULL,
            OPEN_EXISTING, 0, NULL);
        if (hDev == INVALID_HANDLE_VALUE)
            break;

        if (DeviceIoControl(hDev, IOCTL_DISK_GROW_PARTITION,
            &disk_grow_partition,
            sizeof disk_grow_partition, NULL, 0, &dw,
            NULL))
            SetLastError(NO_ERROR);

        break;
    }

    default:
        hDev = CreateFile(cBuf, GENERIC_READ | GENERIC_WRITE,
            FILE_SHARE_READ | FILE_SHARE_WRITE, NULL,
            OPEN_EXISTING, 0, NULL);
        if (hDev == INVALID_HANDLE_VALUE)
            break;

        FlushFileBuffers(hDev);
    }

    fputs("\r\n", stderr);
    win_perror(cBuf + 4);

    return 0;
}

POBJECT_NAME_INFORMATION
GetNativeTargetName(LPCWSTR name)
{
    HANDLE h;
    NTSTATUS status;
    POBJECT_NAME_INFORMATION object_name;
    ULONG res;
    LPVOID new_mem;

    h = CreateFile(name,
        0,
        FILE_SHARE_READ |
        FILE_SHARE_WRITE |
        FILE_SHARE_DELETE,
        NULL,
        OPEN_EXISTING,
        FILE_FLAG_BACKUP_SEMANTICS,
        NULL);

    if (h == INVALID_HANDLE_VALUE)
    {
        return NULL;
    }

    object_name = (POBJECT_NAME_INFORMATION)HeapAlloc(GetProcessHeap(),
        HEAP_GENERATE_EXCEPTIONS, UNICODE_STRING_MAX_BYTES);
    status = NtQueryObject(h, ObjectNameInformation, object_name,
        UNICODE_STRING_MAX_BYTES, &res);

    CloseHandle(h);

    if (!NT_SUCCESS(status))
    {
        HeapFree(GetProcessHeap(), 0, object_name);
        SetLastError(RtlNtStatusToDosError(status));
        return NULL;
    }

    new_mem = HeapReAlloc(GetProcessHeap(), HEAP_GENERATE_EXCEPTIONS,
        object_name, res);

    if (new_mem != NULL)
    {
        object_name = new_mem;
    }

    return object_name;
}

LONG
WINAPI
ExceptionFilter(LPEXCEPTION_POINTERS ExceptionInfo)
{
    LPSTR MsgBuf = NULL;
    DWORD i;

    if (FormatMessageA(78 |
        FORMAT_MESSAGE_ALLOCATE_BUFFER |
        FORMAT_MESSAGE_FROM_HMODULE |
        FORMAT_MESSAGE_IGNORE_INSERTS,
        GetModuleHandleA("ntdll.dll"),
        ExceptionInfo->ExceptionRecord->ExceptionCode, 0,
        (LPSTR)&MsgBuf, 0, NULL))
    {
        CharToOemA(MsgBuf, MsgBuf);
    }
    else
    {
        MsgBuf = NULL;
    }

    if (MsgBuf != NULL)
    {
        fprintf(stderr, "\n"
            "%s\n", MsgBuf);

        LocalFree(MsgBuf);
    }

    fprintf(stderr,
        "\n"
        "Fatal error - unhandled exception.\n"
        "\n"
        "Exception 0x%X at address 0x%p\n",
        ExceptionInfo->ExceptionRecord->ExceptionCode,
        ExceptionInfo->ExceptionRecord->ExceptionAddress);

    for (i = 0;
        i < ExceptionInfo->ExceptionRecord->NumberParameters;
        i++)
    {
        fprintf(stderr,
            "Parameter %u: 0x%p\n",
            i + 1,
            ExceptionInfo->ExceptionRecord->ExceptionInformation[i]);
    }

    flushall();
    ExitProcess((UINT)-1);
}

// This is enough startup code for this program if compiled to use the DLL CRT.
#if !defined(_DEBUG) && !defined(DEBUG) && _MSC_VER < 1900
extern "C" int
wmainCRTStartup()
{
    int argc = 0;
    LPWSTR *argv = CommandLineToArgvW(GetCommandLine(), &argc);
    if (argv == NULL)
    {
#ifndef _M_ARM
        MessageBoxA(NULL, "This program requires Windows NT.",
            "Incompatible Windows version", MB_ICONERROR);
#endif
        ExitProcess((UINT)-1);
    }

    exit(wmain(argc, argv));
}
#endif
