/************************************************************
 //
 //	Program: first.c (kapitel 3,exempel 1)
 //
 //	Beskrivning:
 //	 Ett frsta, minimalt Windows program. Visar ett
 //  tomt fnster utan funktionalitet frutom generell
 //	 hantering av fnstret.
 //
 //	Plattformar: Win16, Win32
 //
 //	(c) Studentlitteratur, Hans-Erik Eriksson
 //	frn boken "Programutveckling fr Windows och Windows NT"
 //
 ************************************************************/

// Inkluderingsfiler
#define WIN32_LEAN_AND_MEAN
#include <windows.h>

#pragma comment(lib, "kernel32.lib")
#pragma comment(lib, "user32.lib")
#pragma comment(lib, "gdi32.lib")

#pragma comment(linker, "/subsystem:windows")
#pragma comment(linker, "/entry:startup")

HBRUSH hBg;

// Fnsterprocedur fr huvudfnster
LRESULT CALLBACK
WindowProc(HWND hWnd,		// Fnster handtag
	   UINT msg,		// Meddelande
	   WPARAM wParam,	// Parameter (w)
	   LPARAM lParam)	// Parameter (l)
{
  switch (msg)
    {
    case WM_DESTROY:
      {
	PostQuitMessage(0);
	return 0;
      }
    default:
      return DefWindowProc(hWnd, msg, wParam, lParam);
    }
}

LPSTR errormsg()
{
  LPSTR lpBuf = NULL;
  if (FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM |
		    FORMAT_MESSAGE_ALLOCATE_BUFFER, NULL, GetLastError(), 0,
		    (LPSTR) &lpBuf, 0, NULL))
    return lpBuf;
  else
    return NULL;
}

// Registrering av fnsterklass
void
DoRegisterClasses()
{
  WNDCLASS wndclass;
  HINSTANCE hInstance = GetModuleHandle(NULL);
  HBITMAP hBitMap;

  hBitMap = LoadBitmap(hInstance, "IDB_BGBITMAP");
  if (hBitMap == NULL)
    {
      MessageBox(NULL, errormsg(), "LoadBitmap()", MB_ICONSTOP | MB_OK);
      return;
    }

  hBg = CreatePatternBrush(hBitMap);
  if (hBg == NULL)
    {
      MessageBox(NULL, errormsg(), "CreateBrushIndirect()",
		 MB_ICONSTOP | MB_OK);
      return;
    }

  wndclass.lpszClassName = "APPLIKATION";	// Klassens namn
  wndclass.hInstance = hInstance;	// Programinstans
  wndclass.lpfnWndProc = WindowProc;	// Fnsterprocedur
  wndclass.hCursor = LoadCursor(NULL, IDC_ARROW);	// Musmarkr
  wndclass.hIcon = LoadIcon(hInstance, "IDI_FIRSTICON");	// Ikon
  wndclass.lpszMenuName = NULL;	// Meny
  wndclass.hbrBackground = (HBRUSH) hBg;	// Bakgrundsfrg
  wndclass.style = 0L;		// Fnsterklass stil
  wndclass.cbClsExtra = 0;	// Extra bytes fr klass
  wndclass.cbWndExtra = 0;	// Extra bytes fr fnster

  // Registrera klass
  RegisterClass(&wndclass);
}

// EXE-filens startkod - registrera fnsterklass, skapa fnster och
// initiera meddelande loop
int
startup()
{
  HWND hWnd;
  MSG msg;
  STARTUPINFO si;
  GetStartupInfo(&si);

  // Registrera fnsterklass
  DoRegisterClasses();

  // Skapa fnster
  hWnd = CreateWindow("APPLIKATION",	// Klassnamn
		      "Programutveckling fr Windows och Windows NT",	// Fnster titel
		      WS_OVERLAPPEDWINDOW,	// Fnster stil
		      CW_USEDEFAULT,	// X-punkt vre vnstra hrn
		      CW_USEDEFAULT,	// Y-punkt vre vnstra hrn
		      CW_USEDEFAULT,	// cx (bredd)
		      CW_USEDEFAULT,	// cy (hjd)
		      NULL,	// Frlderfnster
		      NULL,	// Meny
		      GetModuleHandle(NULL),	// Programinstans
		      NULL);	// Data pekare

  if (!hWnd)
    ExitProcess(GetLastError());

  // Visa fnster
  ShowWindow(hWnd, si.wShowWindow);

  // Meddelandeloop som hmtar meddelanden
  while (GetMessage(&msg, 0, 0, 0) > 0)
    {
      TranslateMessage(&msg);
      DispatchMessage(&msg);
    }

  ExitProcess(0);
}
