#define WIN32_LEAN_AND_MEAN
#define BUILD_WINDOWS

#ifdef UNICODE
#undef UNICODE
#endif
#ifdef _UNICODE
#undef _UNICODE
#endif

#include <stdio.h>

#include <winstrct.h>
#include <winternl.h>

#pragma comment(lib, "kernel32.lib")
#pragma comment(lib, "user32.lib")

typedef NTSTATUS WINAPI fRtlGetVersion(
    __inout __deref PRTL_OSVERSIONINFOW lpVersionInformation);

typedef fRtlGetVersion *pfRtlGetVersion;

BOOL
InternalGetVersion(LPOSVERSIONINFO lpVersionInformation)
{
    char *msg;
    NTSTATUS status;
    OSVERSIONINFOW ntoi = { sizeof(ntoi) };
    pfRtlGetVersion fpRtlGetVersion = (pfRtlGetVersion)
        GetProcAddress(GetModuleHandle("ntdll.dll"), "RtlGetVersion");

    if (fpRtlGetVersion == NULL)
    {
        if (!GetVersionEx(lpVersionInformation))
        {
            msg = mprintf("Error %1!#x!", GetLastError());
            MessageBox(NULL, msg, "Error", MB_ICONSTOP);
            return FALSE;
        }

        return TRUE;
    }

    status = fpRtlGetVersion(&ntoi);

    if (!NT_SUCCESS(status))
    {
        msg = mprintf("Error %1!#x!", status);
        MessageBox(NULL, msg, "Error", MB_ICONSTOP);
        return FALSE;
    }

    memcpy(&lpVersionInformation->dwOSVersionInfoSize + 1,
        &ntoi.dwOSVersionInfoSize + 1,
        FIELD_OFFSET(OSVERSIONINFOW, szCSDVersion) -
        sizeof(ntoi.dwOSVersionInfoSize));

    WideCharToMultiByte(GetACP(), 0, ntoi.szCSDVersion, -1,
        lpVersionInformation->szCSDVersion,
        _countof(lpVersionInformation->szCSDVersion), NULL, NULL);

    return TRUE;
}

int WINAPI
WinMain(HINSTANCE hCurInstance, HINSTANCE hPrev, LPSTR lpCmdLine, int nCmdShow)
{
    OSVERSIONINFO oi = { sizeof(oi) };

    char *osname = NULL;
    char *msg = NULL;

    if (!InternalGetVersion(&oi))
    {
        return 1;
    }

    switch (oi.dwPlatformId)
    {
    case VER_PLATFORM_WIN32s:
        osname = "Win32s";
        break;

    case VER_PLATFORM_WIN32_WINDOWS:
        if (oi.dwMinorVersion < 10)
            osname = "Windows 95";
        else if (oi.dwMinorVersion > 10)
            osname = "Windows ME";
        else
            osname = "Windows 98";
        break;

    case VER_PLATFORM_WIN32_NT:
        if (oi.dwMajorVersion < 5)
            osname = "Windows NT";
        else if (oi.dwMajorVersion == 5)
        {
            if (oi.dwMinorVersion < 1)
                osname = "Windows 2000";
            else if (oi.dwMinorVersion == 1)
                osname = "Windows XP";
            else if (oi.dwMinorVersion == 2)
                osname = "Windows Server 2003";
        }
        else if (oi.dwMajorVersion == 6)
        {
            if (oi.dwMinorVersion < 1)
                osname = "Windows Vista or Server 2008";
            else if (oi.dwMinorVersion == 1)
                osname = "Windows 7 or Server 2008 R2";
            else if (oi.dwMinorVersion == 2)
                osname = "Windows 8 or Server 2012";
            else if (oi.dwMinorVersion == 3)
                osname = "Windows 8.1 or Server 2012 R2";
        }
        else if (oi.dwMajorVersion == 10)
        {
            if (oi.dwBuildNumber < 22000)
                osname = "Windows 10";
			else
				osname = "Windows 11";
        }

        if (osname == NULL)
            osname = "Windows";

        break;

    default:
        osname = "Unknown OS";
    }

    msg = mprintf("%1 version %2!u!.%3!u!%nBuild %4!u! %5",
        osname,
        oi.dwMajorVersion,
        oi.dwMinorVersion,
        oi.dwPlatformId == VER_PLATFORM_WIN32_WINDOWS ?
        oi.dwBuildNumber & 0x0000FFFF : oi.dwBuildNumber,
        oi.szCSDVersion);

    if (msg == NULL)
        msg = win_error;

    MessageBox(NULL, msg, "Win32 version", MB_ICONINFORMATION);
    LocalFree(msg);

    return 0;
}


// This is enough startup code for this program if compiled to use the DLL CRT.
#if !defined(_DEBUG) && !defined(DEBUG) && _MSC_VER < 1900
int
WinMainCRTStartup()
{
    ExitProcess(WinMain(NULL, NULL, NULL, 0));
}
#endif
