#define WIN32_LEAN_AND_MEAN
#define BUILD_WINDOWS

#ifdef UNICODE
#undef UNICODE
#endif
#ifdef _UNICODE
#undef _UNICODE
#endif

#include <stdio.h>

#include <windows.h>
#include <winbase.h>
#include <winuser.h>
#include <winternl.h>

#pragma comment(lib, "kernel32")

#if defined(_DLL) && defined(_M_IX86)
#pragma comment(lib, "nullcrt")
#endif

typedef NTSTATUS WINAPI fRtlGetVersion(
    __inout __deref PRTL_OSVERSIONINFOW lpVersionInformation);

typedef fRtlGetVersion *pfRtlGetVersion;

BOOL
InternalGetVersion(LPOSVERSIONINFO lpVersionInformation)
{
    NTSTATUS status;
    OSVERSIONINFOW ntoi = { sizeof(ntoi) };
    pfRtlGetVersion fpRtlGetVersion = (pfRtlGetVersion)
        GetProcAddress(GetModuleHandle("ntdll.dll"), "RtlGetVersion");

    if (fpRtlGetVersion == NULL)
    {
        if (!GetVersionEx(lpVersionInformation))
        {
            fprintf(stderr, "Error %#x\n", GetLastError());
            return FALSE;
        }

        return TRUE;
    }

    status = fpRtlGetVersion(&ntoi);

    if (!NT_SUCCESS(status))
    {
        fprintf(stderr, "Error %#x\n", status);
        return FALSE;
    }

    memcpy(&lpVersionInformation->dwOSVersionInfoSize + 1,
        &ntoi.dwOSVersionInfoSize + 1,
        FIELD_OFFSET(OSVERSIONINFOW, szCSDVersion) -
        sizeof(ntoi.dwOSVersionInfoSize));

    WideCharToMultiByte(GetOEMCP(), 0, ntoi.szCSDVersion, -1,
        lpVersionInformation->szCSDVersion,
        sizeof(lpVersionInformation->szCSDVersion) /
        sizeof(*lpVersionInformation->szCSDVersion), NULL, NULL);

    return TRUE;
}

int main()
{
    OSVERSIONINFO oi = { sizeof(oi) };

    char *osname = NULL;

    if (!InternalGetVersion(&oi))
    {
        return 1;
    }

    switch (oi.dwPlatformId)
    {
    case VER_PLATFORM_WIN32s:
        osname = "Win32s";
        break;

    case VER_PLATFORM_WIN32_WINDOWS:
        if (oi.dwMinorVersion < 10)
            osname = "Windows 95";
        else if (oi.dwMinorVersion > 10)
            osname = "Windows ME";
        else
            osname = "Windows 98";
        break;

    case VER_PLATFORM_WIN32_NT:
        if (oi.dwMajorVersion < 5)
            osname = "Windows NT";
        else if (oi.dwMajorVersion == 5)
        {
            if (oi.dwMinorVersion < 1)
                osname = "Windows 2000";
            else if (oi.dwMinorVersion == 1)
                osname = "Windows XP";
            else if (oi.dwMinorVersion == 2)
                osname = "Windows Server 2003";
        }
        else if (oi.dwMajorVersion == 6)
        {
            if (oi.dwMinorVersion < 1)
                osname = "Windows Vista or Server 2008";
            else if (oi.dwMinorVersion == 1)
                osname = "Windows 7 or Server 2008 R2";
            else if (oi.dwMinorVersion == 2)
                osname = "Windows 8 or Server 2012";
            else if (oi.dwMinorVersion == 3)
                osname = "Windows 8.1 or Server 2012 R2";
        }
        else if (oi.dwMajorVersion == 10)
        {
            if (oi.dwBuildNumber < 22000)
                osname = "Windows 10";
			else
				osname = "Windows 11";
        }

        if (osname == NULL)
            osname = "Windows";

        break;

    default:
        osname = "Unknown OS";
    }

    printf("%s version %u.%u\r\nBuild %u %s\n",
        osname,
        oi.dwMajorVersion,
        oi.dwMinorVersion,
        oi.dwPlatformId == VER_PLATFORM_WIN32_WINDOWS ?
        oi.dwBuildNumber & 0x0000FFFF : oi.dwBuildNumber,
        oi.szCSDVersion);

    return 0;
}
