#define UNICODE
#define _UNICODE
#define WIN32_LEAN_AND_MEAN
#include <winstrct.h>
#include <shellapi.h>
#include <wfind.h>
#include <wntsecur.h>
#include <stdlib.h>
#include <aclapi.h>

#pragma comment(lib, "kernel32.lib")
#pragma comment(lib, "shell32.lib")
#pragma comment(lib, "advapi32.lib")
#ifdef _M_IX86
#pragma comment(lib, "crtdll.lib")
#endif

#define si (OWNER_SECURITY_INFORMATION|GROUP_SECURITY_INFORMATION|\
   DACL_SECURITY_INFORMATION|SACL_SECURITY_INFORMATION)

void *pSecurityBuffer = NULL;
DWORD dwSecurityBufferSize = 0;
bool takeown = false;
BYTE userSID[260] = { 0 };

bool ReadSecurity(LPCWSTR);
bool WriteSecurity(LPCWSTR);

enum XError
{
    XE_NOERROR,
    XE_CANCELLED,
    XE_COPYFILE,
    XE_CREATEDIR,
    XE_SECURITYREAD,
    XE_SECURITYWRITE,
    XE_CHDIR,
    XE_CHDIRBACK,
    XE_NOT_WINDOWSNT,
    XE_NOT_ENOUGH_MEMORY
};

void
status_exit(XError XE)
{
    const char *errmsg = NULL;
    switch (XE)
    {
    case XE_CANCELLED:
        errmsg = "\r\nCopyACLs cancelled.\r\n";
        break;
    case XE_SECURITYREAD:
        errmsg = "\r\nRead of security descriptor failed. CopyACLs aborted.\r\n";
        break;
    case XE_SECURITYWRITE:
        errmsg = "\r\nFile security descriptor change failed. "
            "CopyACLs aborted.\r\n";
        break;
    case XE_NOT_WINDOWSNT:
        MessageBoxA(NULL, "This program requires Windows NT.", "CopyACLs",
            MB_ICONERROR);
        exit(XE);
    }

    if (errmsg)
        fputs(errmsg, stderr);

    exit(XE);
}

int
wmain(int argc, LPWSTR *argv)
{
    if (argc < 3)
    {
        puts(
            "Usage:\n"
            "CopyACLs source target1 [target2 ...]\n"
            "CopyACLs /restore=sourcedumpfile target1 [target2 ...]\n"
            "CopyACLs source /dump=targetdumpfile1 [/dump=targetdumpfile2 ...]\n"
            "CopyACLs /takeown accountname target1 [target2 ...]\n");
        return 0;
    }

    SetOemPrintFLineLength(GetStdHandle(STD_ERROR_HANDLE));

    HANDLE hToken;
    struct
    {
        TOKEN_PRIVILEGES TP;
        TOKEN_PRIVILEGES Privileges[5];
    } TP;
    if (OpenProcessToken(GetCurrentProcess(), TOKEN_ADJUST_PRIVILEGES, &hToken))
    {
        TP.TP.PrivilegeCount = 5;
        LookupPrivilegeValue(L"", SE_SECURITY_NAME, &(TP.TP.Privileges[0].Luid));
        LookupPrivilegeValue(L"", SE_TAKE_OWNERSHIP_NAME, &(TP.TP.Privileges[1].Luid));
        LookupPrivilegeValue(L"", SE_BACKUP_NAME, &(TP.TP.Privileges[2].Luid));
        LookupPrivilegeValue(L"", SE_RESTORE_NAME, &(TP.TP.Privileges[3].Luid));
        LookupPrivilegeValue(L"", SE_TCB_NAME, &(TP.TP.Privileges[4].Luid));
        LookupPrivilegeValue(L"", SE_IMPERSONATE_NAME, &(TP.TP.Privileges[5].Luid));
        TP.TP.Privileges[0].Attributes =
            TP.TP.Privileges[1].Attributes =
            TP.TP.Privileges[2].Attributes =
            TP.TP.Privileges[3].Attributes =
            TP.TP.Privileges[4].Attributes =
            TP.TP.Privileges[5].Attributes = SE_PRIVILEGE_ENABLED;

        AdjustTokenPrivileges(hToken, FALSE, &TP.TP, NULL, NULL, NULL);
    }

    if (argc >= 4 && _wcsicmp(argv[1], L"/takeown") == 0)
    {
        argc--;
        argv++;

        takeown = true;

        SID_NAME_USE SIDnameuse;
        WCHAR userdomain[260] = L"";
        DWORD userdomainsiz = sizeof userdomain;
        DWORD userSIDsiz = sizeof userSID;

        if (_wcsicmp(argv[1], L"Administrators") == 0)
        {
            CreateWellKnownSid(WinBuiltinAdministratorsSid, NULL, userSID, &userSIDsiz);
        }
        else if (_wcsicmp(argv[1], L"SYSTEM") == 0)
        {
            CreateWellKnownSid(WinLocalSystemSid, NULL, userSID, &userSIDsiz);
        }
        else if (!LookupAccountName(NULL, argv[1], &userSID, &userSIDsiz,
            userdomain, &userdomainsiz, &SIDnameuse))
        {
            win_perror(argv[1]);
            return 1;
        }

        argc--;
        argv++;

        while (argc > 1)
        {
            SE_OBJECT_TYPE object_type;

            if (_wcsnicmp(argv[1], L"HKEY_", 5) == 0)
            {
                object_type = SE_REGISTRY_KEY;
                argv[1] += 5;
            }
            else
            {
                object_type = SE_FILE_OBJECT;
            }

            DWORD dwerr = SetNamedSecurityInfoW(argv[1], object_type, OWNER_SECURITY_INFORMATION, &userSID, NULL, NULL, NULL);

            if (dwerr != NO_ERROR)
            {
                SetLastError(dwerr);
                win_perror(argv[1]);
            }

            argc--;
            argv++;
        }

        return 0;
    }

    if (ReadSecurity(argv[1]))
    {
        while (--argc > 1)
        {
            WriteSecurity((++argv)[1]);
        }

        printf("\nCopyACLs done. Security descriptor size: %u bytes.\n",
            dwSecurityBufferSize);
    }
    else
    {
        win_perrorA("Error reading security descriptor");
    }

    return XE_NOERROR;
}

bool
RecognizeKeyName(LPCWSTR rootkey, LPCWSTR shortname, LPCWSTR longname, HKEY hkroot, PHKEY phkroot)
{
    if (_wcsicmp(rootkey, shortname) == 0 || _wcsicmp(rootkey, longname) == 0)
    {
        *phkroot = hkroot;
        return true;
    }
    else
    {
        return false;
    }
}

HKEY
OpenKeyPath(LPCWSTR path, REGSAM sam)
{
    LPWSTR rootkey = _wcsdup(path);

    if (rootkey == NULL)
    {
        return NULL;
    }

    wcstok(rootkey, L"\\");
    
    LPCWSTR subkey = wcstok(NULL, L"");

    if (subkey != NULL)
    {
        subkey = path + (subkey - rootkey);
    }

    HKEY hkey = NULL;

    RecognizeKeyName(rootkey, L"HKCR", L"HKEY_CLASSES_ROOT", HKEY_CLASSES_ROOT, &hkey) ||
        RecognizeKeyName(rootkey, L"HKCU", L"HKEY_CURRENT_USER", HKEY_CURRENT_USER, &hkey) ||
        RecognizeKeyName(rootkey, L"HKLM", L"HKEY_LOCAL_MACHINE", HKEY_LOCAL_MACHINE, &hkey) ||
        RecognizeKeyName(rootkey, L"HKU", L"HKEY_USERS", HKEY_USERS, &hkey) ||
        RecognizeKeyName(rootkey, L"HKPD", L"HKEY_PERFORMANCE_DATA", HKEY_PERFORMANCE_DATA, &hkey) ||
        RecognizeKeyName(rootkey, L"HKPT", L"HKEY_PERFORMANCE_TEXT", HKEY_PERFORMANCE_TEXT, &hkey) ||
        RecognizeKeyName(rootkey, L"HKPNT", L"HKEY_PERFORMANCE_NLSTEXT", HKEY_PERFORMANCE_NLSTEXT, &hkey) ||
        RecognizeKeyName(rootkey, L"HKCC", L"HKEY_CURRENT_CONFIG", HKEY_CURRENT_CONFIG, &hkey) ||
        RecognizeKeyName(rootkey, L"HKDD", L"HKEY_DYN_DATA", HKEY_DYN_DATA, &hkey) ||
        RecognizeKeyName(rootkey, L"HKCULS", L"HKEY_CURRENT_USER_LOCAL_SETTINGS", HKEY_CURRENT_USER_LOCAL_SETTINGS, &hkey);

    free(rootkey);

    if (hkey == NULL)
    {
        SetLastError(ERROR_REGISTRY_IO_FAILED);
        return NULL;
    }

    DWORD dwerr = RegOpenKeyEx(hkey, subkey, 0, sam, &hkey);

    if (dwerr != NO_ERROR)
    {
        SetLastError(dwerr);
        return NULL;
    }

    return hkey;
}

bool
ReadKeySecurity(LPCWSTR source)
{
    HKEY hkey = OpenKeyPath(source, KEY_READ | ACCESS_SYSTEM_SECURITY);

    if (hkey == NULL)
    {
        return false;
    }

    DWORD dwSDSizeNeeded = 0;
    while (!((WSecurityDescriptor*)pSecurityBuffer)->
        FromKey(hkey, si, dwSecurityBufferSize, &dwSDSizeNeeded))
        switch (GetLastError())
        {
        case ERROR_INSUFFICIENT_BUFFER:
            if (pSecurityBuffer)
                hfree(pSecurityBuffer);

            pSecurityBuffer = halloc(dwSDSizeNeeded);
            if (pSecurityBuffer == NULL)
                status_exit(XE_NOT_ENOUGH_MEMORY);

            dwSecurityBufferSize = dwSDSizeNeeded;

            continue;

        case ERROR_CALL_NOT_IMPLEMENTED:
            RegCloseKey(hkey);
            status_exit(XE_NOT_WINDOWSNT);

        default:
            WErrMsgA errmsg;
            oem_printf(stderr,
                "Error reading security descriptor from '%1!ws!': %2%%n",
                source, (LPCSTR)errmsg);

            RegCloseKey(hkey);
            status_exit(XE_SECURITYREAD);
        }

    RegCloseKey(hkey);
    return true;
}

bool
ReadDumpFile(LPCWSTR source)
{
    HANDLE h = CreateFile(source, GENERIC_READ, FILE_SHARE_READ | FILE_SHARE_DELETE, NULL, OPEN_EXISTING, 0, NULL);
    if (h == INVALID_HANDLE_VALUE)
    {
        WErrMsgA errmsg;
        oem_printf(stderr,
            "Error reading security descriptor from '%1!ws!': %2%%n",
            source, (LPCSTR)errmsg);
        status_exit(XE_SECURITYREAD);
    }

    ULARGE_INTEGER file_size;
    file_size.LowPart = GetFileSize(h, &file_size.HighPart);
    if (file_size.LowPart == INVALID_FILE_SIZE && GetLastError() != NO_ERROR)
    {
        WErrMsgA errmsg;
        oem_printf(stderr,
            "Error reading security descriptor from '%1!ws!': %2%%n",
            source, (LPCSTR)errmsg);

        CloseHandle(h);

        status_exit(XE_SECURITYREAD);
    }

    if (file_size.HighPart != 0)
    {
        oem_printf(stderr,
            "File too large: '%1!ws!'%%n", source);

        CloseHandle(h);

        status_exit(XE_SECURITYREAD);
    }

    if (dwSecurityBufferSize < file_size.LowPart)
    {
        if (pSecurityBuffer)
            hfree(pSecurityBuffer);

        pSecurityBuffer = halloc(file_size.LowPart);
        if (pSecurityBuffer == NULL)
            status_exit(XE_NOT_ENOUGH_MEMORY);

        dwSecurityBufferSize = file_size.LowPart;
    }

    DWORD dw;
    if (!ReadFile(h, pSecurityBuffer, file_size.LowPart, &dw, NULL))
    {
        WErrMsgA errmsg;
        oem_printf(stderr,
            "Error reading security descriptor from '%1!ws!': %2%%n",
            source, (LPCSTR)errmsg);

        CloseHandle(h);

        status_exit(XE_SECURITYREAD);
    }

    dwSecurityBufferSize = dw;

    CloseHandle(h);

    return true;
}

bool
ReadSecurity(LPCWSTR source)
{
    if (_wcsnicmp(source, L"HK", 2) == 0)
    {
        return ReadKeySecurity(source);
    }
    else if (_wcsnicmp(source, L"/restore=", 9) == 0)
    {
        return ReadDumpFile(source + 9);
    }

    DWORD dwSDSizeNeeded = 0;
    while (!((WSecurityDescriptor*)pSecurityBuffer)->
        FromFile(source, si, dwSecurityBufferSize, &dwSDSizeNeeded))
    {
        switch (GetLastError())
        {
        case ERROR_INSUFFICIENT_BUFFER:
            if (pSecurityBuffer)
                hfree(pSecurityBuffer);

            pSecurityBuffer = halloc(dwSDSizeNeeded);
            if (pSecurityBuffer == NULL)
                status_exit(XE_NOT_ENOUGH_MEMORY);

            dwSecurityBufferSize = dwSDSizeNeeded;

            continue;

        case ERROR_CALL_NOT_IMPLEMENTED:
            status_exit(XE_NOT_WINDOWSNT);

        default:
            WErrMsgA errmsg;
            oem_printf(stderr,
                "Error reading security descriptor from '%1!ws!': %2%%n",
                source, (LPCSTR)errmsg);
            status_exit(XE_SECURITYREAD);
        }
    }

    return true;
}

bool
WriteKeySecurity(LPCWSTR target)
{
    HKEY hkey = OpenKeyPath(target, WRITE_DAC | WRITE_OWNER | ACCESS_SYSTEM_SECURITY);

    if (hkey == NULL)
    {
        return false;
    }

    if (((WSecurityDescriptor*)pSecurityBuffer)->ToKey(hkey, si))
    {
        RegCloseKey(hkey);
        return true;
    }

    WErrMsgA errmsg;
    oem_printf(stderr,
        "Error writing security descriptor to '%1!ws!': %2%%n",
        target, (LPCSTR)errmsg);

    RegCloseKey(hkey);
    return false;
}

bool
WriteDumpFile(LPCWSTR target)
{
    HANDLE h = CreateFile(target, GENERIC_WRITE, FILE_SHARE_DELETE, NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
    if (h == INVALID_HANDLE_VALUE)
    {
        WErrMsgA errmsg;
        oem_printf(stderr,
            "Error creating file '%1!ws!': %2%%n",
            target, (LPCSTR)errmsg);
        status_exit(XE_SECURITYREAD);
    }

    DWORD dw;
    if (!WriteFile(h, pSecurityBuffer, dwSecurityBufferSize, &dw, NULL))
    {
        WErrMsgA errmsg;
        oem_printf(stderr,
            "Error writing security descriptor to '%1!ws!': %2%%n",
            target, (LPCSTR)errmsg);

        CloseHandle(h);

        return false;
    }

    CloseHandle(h);

    return true;
}

bool
WriteSecurity(LPCWSTR target)
{
    if (_wcsnicmp(target, L"HK", 2) == 0)
    {
        if (WriteKeySecurity(target))
            return true;
    }
    else if (_wcsnicmp(target, L"/dump=", 6) == 0)
    {
        if (WriteDumpFile(target + 6))
            return true;
    }
    else
    {
        if (((WSecurityDescriptor*)pSecurityBuffer)->ToFile(target, si))
            return true;
    }

    WErrMsgA errmsg;
    oem_printf(stderr,
        "Error writing security descriptor to '%1!ws!': %2%%n",
        target, (LPCSTR)errmsg);
    return false;
}

#if !defined(_DEBUG) && !defined(DEBUG) && _MSC_VER < 1900
extern "C" int
wmainCRTStartup()
{
    int argc = 0;
    LPWSTR *argv = CommandLineToArgvW(GetCommandLine(), &argc);
    if (argv == NULL)
        status_exit(XE_NOT_WINDOWSNT);
    else
        exit(wmain(argc, argv));
}
#endif
