#ifndef WIN32_LEAN_AND_MEAN
#define WIN32_LEAN_AND_MEAN
#endif
#ifndef UNICODE
#define UNICODE
#endif
#ifndef _UNICODE
#define _UNICODE
#endif

#include <stdlib.h>

#include <winstrct.h>
#include <shellapi.h>
#include <wtime.h>

#ifndef FILE_ATTRIBUTE_INTEGRITY_STREAM
#define FILE_ATTRIBUTE_INTEGRITY_STREAM     0x00008000
#endif
#ifndef FILE_ATTRIBUTE_VIRTUAL
#define FILE_ATTRIBUTE_VIRTUAL              0x00010000
#endif
#ifndef FILE_ATTRIBUTE_NO_SCRUB_DATA
#define FILE_ATTRIBUTE_NO_SCRUB_DATA        0x00020000
#endif
#ifndef FILE_ATTRIBUTE_EA
#define FILE_ATTRIBUTE_EA                   0x00040000
#endif
#ifndef FILE_ATTRIBUTE_PINNED
#define FILE_ATTRIBUTE_PINNED                    0x00080000  
#endif
#ifndef FILE_ATTRIBUTE_UNPINNED
#define FILE_ATTRIBUTE_UNPINNED                    0x00100000  
#endif
#ifndef FILE_ATTRIBUTE_RECALL_ON_OPEN
#define FILE_ATTRIBUTE_RECALL_ON_OPEN                    0x00040000  
#endif
#ifndef FILE_ATTRIBUTE_RECALL_ON_DATA_ACCESS
#define FILE_ATTRIBUTE_RECALL_ON_DATA_ACCESS             0x00400000  
#endif
#ifndef FILE_ATTRIBUTE_STRICTLY_SEQUENTIAL
#define FILE_ATTRIBUTE_STRICTLY_SEQUENTIAL             0x20000000  
#endif

#if defined(_DLL) && defined(_M_IX86)
#pragma comment(lib, "minwcrt.lib")
#endif

#pragma comment(lib, "shell32.lib")

void testattr(LPDWORD dwAttr, DWORD dwTestAttr, LPCSTR lpMsg)
{
    if (*dwAttr & dwTestAttr)
    {
        printf(" %s", lpMsg);
        *dwAttr &= ~dwTestAttr;
    }
}

void printfattr(LPCSTR lpMsg, DWORD dwAttr)
{
    printf("%s", lpMsg);

    testattr(&dwAttr, FILE_ATTRIBUTE_READONLY, "Read-only");
    testattr(&dwAttr, FILE_ATTRIBUTE_HIDDEN, "Hidden");
    testattr(&dwAttr, FILE_ATTRIBUTE_SYSTEM, "System");
    testattr(&dwAttr, FILE_ATTRIBUTE_DIRECTORY, "Directory");
    testattr(&dwAttr, FILE_ATTRIBUTE_ARCHIVE, "Archive");
    testattr(&dwAttr, FILE_ATTRIBUTE_DEVICE, "Device");
    testattr(&dwAttr, FILE_ATTRIBUTE_NORMAL, "Normal");
    testattr(&dwAttr, FILE_ATTRIBUTE_TEMPORARY, "Temporary");
    testattr(&dwAttr, FILE_ATTRIBUTE_SPARSE_FILE, "Sparse");
    testattr(&dwAttr, FILE_ATTRIBUTE_REPARSE_POINT, "Reparse point");
    testattr(&dwAttr, FILE_ATTRIBUTE_COMPRESSED, "Compressed");
    testattr(&dwAttr, FILE_ATTRIBUTE_OFFLINE, "Offline");
    testattr(&dwAttr, FILE_ATTRIBUTE_NOT_CONTENT_INDEXED, "Not content indexed");
    testattr(&dwAttr, FILE_ATTRIBUTE_ENCRYPTED, "Encrypted");
    testattr(&dwAttr, FILE_ATTRIBUTE_INTEGRITY_STREAM, "Integrity stream");
    testattr(&dwAttr, FILE_ATTRIBUTE_VIRTUAL, "Virtual");
    testattr(&dwAttr, FILE_ATTRIBUTE_NO_SCRUB_DATA, "No scrub data");
    testattr(&dwAttr, FILE_ATTRIBUTE_EA, "Extended attributes");
    testattr(&dwAttr, FILE_ATTRIBUTE_PINNED, "Pinned");
    testattr(&dwAttr, FILE_ATTRIBUTE_UNPINNED, "Unpinned");
    testattr(&dwAttr, FILE_ATTRIBUTE_RECALL_ON_OPEN, "Recall on open");
    testattr(&dwAttr, FILE_ATTRIBUTE_RECALL_ON_DATA_ACCESS, "Recall on data access");
    testattr(&dwAttr, FILE_ATTRIBUTE_STRICTLY_SEQUENTIAL, "Strictly sequential");

    if (dwAttr != 0)
    {
        printf(" 0x%X", dwAttr);
    }

    puts("");
}

void printdt(const char *msg, FILETIME ft)
{
    WSystemTime st = ft;

    fputs(msg, stdout);

    char cBuf[80] = "";
    if (GetDateFormatA(LOCALE_USER_DEFAULT, DATE_SHORTDATE, &st, NULL, cBuf,
        sizeof cBuf))
        fputs(cBuf, stdout);

    fputc(' ', stdout);

    if (GetTimeFormatA(LOCALE_USER_DEFAULT, 0, &st, NULL, cBuf, sizeof cBuf))
        fputs(cBuf, stdout);

    fputs("\r\n", stdout);
}

void print_info(HANDLE h)
{
    printf("File type: ");

    switch (GetFileType(h))
    {
    case FILE_TYPE_CHAR:
        puts("Character device.");
        break;
    case FILE_TYPE_DISK:
        puts("Disk file.");
        break;
    case FILE_TYPE_PIPE:
        puts("Socket or pipe.");
        break;
    case FILE_TYPE_REMOTE:
        puts("Remote file.");
        break;
    default:
        puts("Unknown.");
    }

    BY_HANDLE_FILE_INFORMATION info;
    if (GetFileInformationByHandle(h, &info))
    {
        printfattr("File attributes:", info.dwFileAttributes);
        if (*(QWORD*)&info.ftCreationTime)
            printdt("Creation time: ", info.ftCreationTime);
        if (*(QWORD*)&info.ftLastAccessTime)
            printdt("Last access time: ", info.ftLastAccessTime);
        if (*(QWORD*)&info.ftLastWriteTime)
            printdt("Last write time: ", info.ftLastWriteTime);
        if (info.dwVolumeSerialNumber)
            printf("Volume serial number: %.4X-%.4X\n",
            (DWORD)HIWORD(info.dwVolumeSerialNumber),
            (DWORD)LOWORD(info.dwVolumeSerialNumber));

        ULARGE_INTEGER FileSize;
        FileSize.HighPart = info.nFileSizeHigh;
        FileSize.LowPart = info.nFileSizeLow;
        printf("File size: %.4g %s\n",
            TO_h(FileSize.QuadPart), TO_p(FileSize.QuadPart));

        printf("Number of file system links: %i\n", (int)info.nNumberOfLinks);

        ULARGE_INTEGER FileIndex;
        FileIndex.HighPart = info.nFileIndexHigh;
        FileIndex.LowPart = info.nFileIndexLow;
        printf("Sequence: %u\n", HIWORD(info.nFileIndexHigh));
        printf("Number: %.f\n",
            (double)(FileIndex.QuadPart & 0x0000FFFFFFFFFFFF));
    }
    else
    {
        ULARGE_INTEGER FileSize;
        FileSize.LowPart = GetFileSize(h, &FileSize.HighPart);
        if (FileSize.LowPart == INVALID_FILE_SIZE)
            if (GetLastError() != ERROR_SUCCESS)
            {
                win_perrorA("No information available");
                return;
            }

        printf("File size: %.4g %s\n",
            TO_h(FileSize.QuadPart), TO_p(FileSize.QuadPart));
    }
}

int wmain(int argc, LPWSTR *argv)
{
    if (argc == 1)
    {
        puts("stdin:");
        print_info(GetStdHandle(STD_INPUT_HANDLE));
        return 0;
    }

    while (argc > 1)
    {
        DWORD dwFileAttr = GetFileAttributes(argv[1]);
        if (dwFileAttr == INVALID_FILE_ATTRIBUTES)
            dwFileAttr = 0;

        HANDLE h = CreateFile(argv[1], 0, FILE_SHARE_READ | FILE_SHARE_WRITE, NULL,
            OPEN_EXISTING, FILE_FLAG_BACKUP_SEMANTICS |
            (dwFileAttr & FILE_ATTRIBUTE_REPARSE_POINT ?
        FILE_FLAG_OPEN_REPARSE_POINT : 0), NULL);

        if (h == INVALID_HANDLE_VALUE)
            win_perror(argv[1]);
        else
        {
            LPSTR oem_file = WideToOemAlloc(argv[1]);
            printf("\nFilename: %s\n", oem_file);
            hfree(oem_file);
            print_info(h);
            CloseHandle(h);
        }

        argc--;
        argv++;
    }
}

// This is enough startup code for this program if compiled to use the DLL CRT.
#if !defined(_DEBUG) && !defined(DEBUG) && _MSC_VER < 1900
extern "C" int
wmainCRTStartup()
{
    int argc = 0;
    LPWSTR *argv = CommandLineToArgvW(GetCommandLine(), &argc);
    if (argv == NULL)
    {
#ifndef _M_ARM
        MessageBoxA(NULL, "This program requires Windows NT.",
            "Incompatible Windows version", MB_ICONERROR);
#endif
        ExitProcess((UINT)-1);
    }

    exit(wmain(argc, argv));
}
#endif
