#define UNICODE
#define _UNICODE

#define WIN32_LEAN_AND_MEAN
#define _WIN32_WINNT 0x0500

#ifndef _DLL
#define _DLL
#endif

#include <winstrct.h>
#include <shellapi.h>
#include <sddl.h>
#include <ntdll.h>
#include <minwcrtlib.h>

#pragma comment(lib, "kernel32.lib")
#pragma comment(lib, "user32.lib")
#pragma comment(lib, "shell32.lib")
#pragma comment(lib, "advapi32.lib")
#pragma comment(lib, "ntdll.lib")

#ifdef _DEBUG
#pragma comment(linker, "/nodefaultlib:msvcrtd.lib")
#pragma comment(linker, "/defaultlib:msvcrt.lib")
#endif

BOOL
WINAPI
ParseSid(LPCWSTR SidString, PSID* Sid)
{
    if (wcsncmp(SidString, L"S-1-", 4) != 0)
    {
        SetLastError(ERROR_INVALID_SID);
        return FALSE;
    }

    UCHAR fields = 0;

    LPCWSTR delim = wcschr(SidString + 4, L'-');

    while (delim != NULL)
    {
        fields++;

        if (fields > SID_MAX_SUB_AUTHORITIES)
        {
            SetLastError(ERROR_INVALID_SID);
            return FALSE;
        }

        delim = wcschr(delim + 1, L'-');
    }

    if (fields < 1)
    {
        SetLastError(ERROR_INVALID_SID);
        return FALSE;
    }

    size_t size = FIELD_OFFSET(SID, SubAuthority) + (fields * RTL_FIELD_SIZE(SID, SubAuthority));

    ULONGLONG auth;

    WCHAR chr;

    if (swscanf(SidString + 4, L"%I64u%c", &auth, &chr) != 2 ||
        (auth & 0xFFFF000000000000ULL) != 0 ||
        chr != L'-')
    {
        SetLastError(ERROR_INVALID_SID);
        return FALSE;
    }

    SID_IDENTIFIER_AUTHORITY sid_auth;
    sid_auth.Value[5] = (BYTE)auth;
    sid_auth.Value[4] = (BYTE)(auth >> 8);
    sid_auth.Value[3] = (BYTE)(auth >> 16);
    sid_auth.Value[2] = (BYTE)(auth >> 24);
    sid_auth.Value[1] = (BYTE)(auth >> 32);
    sid_auth.Value[0] = (BYTE)(auth >> 40);

    *Sid = (PISID)LocalAlloc(LMEM_FIXED, size);

    if (*Sid == NULL)
    {
        return FALSE;
    }

    RtlInitializeSid(*Sid, &sid_auth, fields);

    delim = wcschr(wcschr(SidString + 4, L'-'), L'-');

    for (UCHAR field = 0; field < fields; field++)
    {
        delim++;

        *RtlSubAuthoritySid(*Sid, field) = wcstoul(delim, (LPWSTR*)&delim, 0);

        if (field == fields - 1 && *delim == 0)
        {
            break;
        }

        if (*delim != '-')
        {
            LocalFree(*Sid);
            SetLastError(ERROR_INVALID_SID);
            return FALSE;
        }
    }

    return TRUE;
}

int
wmain(int argc, LPWSTR *argv)
{
    SetOemPrintFLineLength(GetStdHandle(STD_ERROR_HANDLE));

    if (argc < 2)
    {
        fputs("Utility to display basic account information about a user account given a user\r\n"
            "name or SID.\r\n"
            "\n"
            "Syntax:\r\n"
            "lookupacc [computer] S-n-n-n\r\n"
            "lookupacc [computer] [domain\\]username\r\n"
            "lookupacc [computer] username[@fqdn]\r\n", stderr);

        return 1;
    }

    if (argc > 3)
    {
        fputs("Too many parameters.\r\n", stderr);
        return 1;
    }

    LPWSTR computername = NULL;
    if (argc > 2)
    {
        computername = argv[1];
        argc--;
        argv++;
    }

    SID_NAME_USE SIDnameuse;
    WCHAR userdomain[260] = L"";
    DWORD userdomainsiz = sizeof userdomain;
    WCHAR username[260] = L"";
    DWORD usernamesiz = sizeof username;
    BYTE userSID[260] = { 0 };
    DWORD userSIDsiz = sizeof userSID;
    LPWSTR wczUsername = L"(unknown)";
    PSID sidLookup = NULL;
    UNICODE_STRING StringSID = { 0 };
    BOOL bConversionResult = FALSE;

    bConversionResult = ParseSid(argv[1], &sidLookup);

    if (bConversionResult)
    {
        if (!LookupAccountSid(computername, sidLookup, username, &usernamesiz,
            userdomain, &userdomainsiz, &SIDnameuse))
        {
            win_perror();
            return 1;
        }

        RtlInitUnicodeString(&StringSID, argv[1]);
        wczUsername = username;
    }
    else
    {
        if (!LookupAccountName(computername, argv[1], &userSID, &userSIDsiz,
            userdomain, &userdomainsiz, &SIDnameuse))
        {
            win_perror();
            return 1;
        }

        RtlConvertSidToUnicodeString(&StringSID, &userSID, TRUE);
        wczUsername = argv[1];
    }

    LPSTR czSIDnameuse;
    switch (SIDnameuse)
    {
    case SidTypeUser:
        czSIDnameuse = "User";
        break;
    case SidTypeGroup:
        czSIDnameuse = "Group";
        break;
    case SidTypeDomain:
        czSIDnameuse = "Domain";
        break;
    case SidTypeAlias:
        czSIDnameuse = "Alias";
        break;
    case SidTypeWellKnownGroup:
        czSIDnameuse = "Well known group";
        break;
    case SidTypeDeletedAccount:
        czSIDnameuse = "Deleted account";
        break;
    case SidTypeInvalid:
        czSIDnameuse = "Invalid";
        break;
    case SidTypeUnknown:
        czSIDnameuse = "Unknown";
        break;
    case SidTypeComputer:
        czSIDnameuse = "Computer";
        break;
    default:
        czSIDnameuse = "Not known";
        break;
    }

    oem_printf(stdout,
        "Account type: %1%%n"
        "User domain: %2!ws!%%n"
        "User name: %3!ws!%%n"
        "SID: %4!wZ!%%n",
        czSIDnameuse,
        userdomain,
        wczUsername,
        &StringSID);

    return 0;
}

// This is enough startup code for this program if compiled to use the DLL CRT.
#if !defined(_DEBUG) && !defined(DEBUG) && _MSC_VER < 1900
extern "C" int
wmainCRTStartup()
{
    int argc = 0;
    LPWSTR *argv = CommandLineToArgvW(GetCommandLine(), &argc);
    if (argv == NULL)
    {
#ifndef _M_ARM
        MessageBoxA(NULL, "This program requires Windows NT.",
            "Incompatible Windows version", MB_ICONERROR);
#endif
        ExitProcess((UINT)-1);
    }

    exit(wmain(argc, argv));
}
#endif
