#include <stdio.h>

#ifdef _WIN32
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <malloc.h>
#else
#include <unistd.h>
#include <stdlib.h>
#define Sleep(n) usleep(n)
#endif

#include "primeser.h"

#pragma comment(lib, "kernel32")

// Den hr funktionen tar fram en matris med alla
// primfaktorer som talet no kan delas upp i. Matrisen
// kommer att termineras med 1. Uppstr ett fel returneras
// false.
bool primefactorexpress( int primefactors[], int no )
{
   // Om talet r mindre n 2 frsker vi inte ens. Vi
   // returnerar en terminerande etta som enda innehll i
   // matrisen.
   if( no < 2 )
   {
      primefactors[0] = 1;
      return true;  
   }

   // Om det r ett jmnt tal delar vi med 2 tills det blir
   // udda. Fr varje tva lgger vi en tva i matrisen och
   // flyttar till nsta element i den.
   while( (no & 1) == 0 )
   {
      *(primefactors++) = 2;

      // Flytta alla bitar ett steg hger ger division med
      // 2.
      no >>= 1;
      Sleep(0);   // Sov lite s att vi inte ter CPU-kraft
   }

   // Om talet nu r mindre n 3 (kan d egentligen bara
   // vara 1 eller negativt) fortstter vi inte. Vi
   // terminerar matrisen hr och returnerar.
   if( no < 3 )
   {
      primefactors[0] = 1;
      return true;
   }

   // Antal udda tal som ska underskas r antalet tal
   // delat med 2.
   int serieslength = no >> 1;

   // Allokera minne fr primtalserien som ska genereras fr
   // att underska om tal r primtal.
   int *primeseries = (int*)alloca(serieslength * sizeof(int));

   // Om alloca()-funktionen returnerade NULL s fanns inte
   // tillrckligt med minne ledigt.
   if( primeseries == NULL )
     {
       fputs("Det finns inte tillrckligt minne.\r\n", stderr);
       return false;
     }

   // Vi brjar p tal 3. Lngden p primtalserien r frn
   // brjan 0.
   int    currentno = 3,
         currentserieslength = 0;
   // verbose-variabeln r true om vi har kommit fram till
   // talet som ska underskas.
   bool    verbose = false;
   // Ta fram max s mnga primtal som fr plats i primtal-
   // serie-matrisen som allokerades.
   for( int counter = 0; counter < serieslength; counter++ )
   {
      // Om vi r p sista talet (det som ska underskas)
      // stter vi variabeln verbose till true.
      verbose = counter + 1 == serieslength;

      if( verbose )
         verbose = true;

      // Anvnd isprimeno()-funktionen fr att ta reda p om
      // talet vi r p r ett primtal.
      if( isprimeno( currentno, primefactors, primeseries, currentserieslength, verbose ) )
      {
         // Om det nu r ett primtal, s lgger vi till det
         // i vr primtalsmatris och kar lngden p den.
         primeseries[currentserieslength++] = currentno;
         // Om vi r p sista talet...
         if( verbose )
         {
            // ...s har vi upptckt att det var ett
            //   primtal. Primfaktorserien blir allts bara
            // talet sjlvt. Vi terminerar ocks med en
            // etta.
            primefactors[0] = no;
            primefactors[1] = 1;
            return true;
         }
      }

      // D gr vi vidare till nsta udda tal.
      currentno += 2;
      Sleep(0);   // Sov lite s att vi inte ter CPU-kraft
   }

   return true;
}

// Den hr funktionen tar reda p om talet no r ett
// primtal med hjlp av primtalserien primeseries med
// lngden currentserieslength. Om verbose r true placeras
// talets primfaktorer i primefactors.
// Funktionen returnerar true om no r ett primtal, annars
// false.
bool isprimeno( int no, int primefactors[], int* &primeseries, const int &currentserieslength, const bool &verbose )
{
   int quote;

   // Vi sker igenom hela primtalserien fr att hitta
   // ngot som talet no kan delas med.
   for( int seekcou = 0; seekcou < currentserieslength; seekcou ++ )
      // Om det hr talet r talet vi r p i matrisen
      if( no == primeseries[seekcou] )
      {
         // S r det givetvis ett primtal
         primefactors[0] = primeseries[seekcou];
         primefactors[1] = 1;
         Sleep(0); // Sov lite s att vi inte ter CPU-kraft
         return true;
      }
      // Om no r delbart med det hr talet (Kvoten sparas i en variabel)
      else if( (quote = no / primeseries[seekcou]) * primeseries[seekcou] == no )
      {
         // Om vi ska gra en primfaktorlista...
         if( verbose )
         {
            // ...s r ju fr det frsta det hr talet en
            // primfaktor. Vi stter sedan en terminering
            // eftersom vi inte vet om det finns fler prim-
            // faktorer.
            primefactors[0] = primeseries[seekcou];
            primefactors[1] = 1;

            // Vi anvnder ocks den hr funktionen fr att
            // ta reda p om kvoten r ett primtal. Vi
            // skickar d en pekare till ett element lngre
            // fram i matrisen primefactors, eftersom den
            // kapslade funktionen ska behandla nsta
            // primfaktor.
            isprimeno( quote, &(primefactors[1]), primeseries, currentserieslength, verbose );
         }
         Sleep(0); // Sov lite s att vi inte ter CPU-kraft
         return false;
      }

   Sleep(0);   // Sov lite s att vi inte ter CPU-kraft
   return true;
}

